import asyncio
import logging
import os
import time
from enum import IntEnum

import configargparse
import cv2 as cv
from autobahn.asyncio.wamp import ApplicationRunner, ApplicationSession
from autobahn.wamp.types import ComponentConfig
from dnnutils import gpslib, utils
from gpxpy.geo import Location
from sanic import response

from . import api_cloud, sensors, truth
from .db import Db
from .kmllib import Kml, ZoneType
from .rest import Authorize, RestApi

log = logging.getLogger(__name__)


class ParkingAngle(IntEnum):
    parallel = 0
    vertical = 1


def initParser():
    parser = configargparse.get_argument_parser()
    parser.add_argument(
        "--config", required=False, is_config_file=True, help="config file path"
    )
    parser.add_argument(
        "--config-save",
        required=False,
        is_write_out_config_file_arg=True,
        help="config file path",
    )

    parser.add_argument("--debug", action="store_true", default=False, help="debug")
    # parser.add_argument('--upload', action='store_true', default=False, help="send info to kiunsys")
    parser.add_argument("--incoming", default="/tmp/json/*.json")
    # parser.add_argument("--processed", default="/tmp/processed")
    parser.add_argument("--ws", default="ws://localhost:8082/ws")
    parser.add_argument("--realm", default="realm1")
    parser.add_argument(
        "--validate-delay",
        type=int,
        default=5,
        help="delay in second to validate plates from two cameras",
    )
    parser.add_argument(
        "--plate-images", default="/tmp/images", help="path to load jpg images plate"
    )
    parser.add_argument(
        "--plates-truth", default="", help="path to load file with plates"
    )
    parser.add_argument("--mev-unit", default="noname", help="unit name")
    parser.add_argument("--flag-gps-fix", default="/tmp/gps-fix", help="gps is valid")
    parser.add_argument(
        "--db-reset", action="store_true", default=False, help="reset database"
    )

    config = parser.parse_args()

    logging.basicConfig(
        format="%(asctime)s [%(levelname)s] %(name)s %(message)s",
        level="DEBUG" if config.debug else "INFO",
    )

    return config


class PollerMev(ApplicationSession, RestApi):
    api: api_cloud.Api  # valorizzato dal login al cloud

    def __init__(self, config):
        ApplicationSession.__init__(self, ComponentConfig(config.realm))
        RestApi.__init__(self, config)
        self.settings = config

        self.db_plates: Db = None
        self.api = None

        now = int(time.time())  # datetime.now().strftime("%Y-%m-%d_%H:%M:%S")
        self.gps = gpslib.Gps(config, filename="{}.gpx".format(now))
        self.kml: Kml = Kml(config, filename="{}.kml".format(now))
        self.save_status = False

        self.db_plates = Db(config.db, reset=config.db_reset)
        self.truth_plates = truth.Truth(config)

        self.gps.initialize()
        self.kml.initialize()
        # self.gps.addCallback(self.kml.addRoutePoint)  # aggiorniamo il file kml con la rotta dalla vettura
        self._counter_gps = 0
        self.gps.addCallback(self.publishLocation)

        # self.oldLocation = {
        #     "latitude": 0,
        #     "longitude": 0,
        #     "speed": 0,
        #     "satellites": 0,
        #     "hdop": 0,
        #     "elevation": 0,
        #     "sensors": 0,
        #     }
        self.oldLocation = {}
        self.newLocation = {}
        
        runner = ApplicationRunner(config.ws, config.realm, extra={})
        asyncio.get_event_loop().run_until_complete(runner.run(self, start_loop=False))

    async def onJoin(self, details):
        log.info("wamp session attached to realm %s", self.settings.realm)

        self.subscribe(self.onPlate, "plate.new")
        self.subscribe(self.onSaveToggle, "dcam.toggle")
        if self.settings.gps == "remote":
            log.info("gps remote: subscribe to gps.remote")
            self.subscribe(self.onGpsRemote, "gps.remote")  # from smartphone
        self.register(self.getStatus, "dcam.status")
        self.register(self.getPlateImage, "plate.image.get")
        self.register(self.getZones, "zones.get")

        asyncio.create_task(self.poll())
        asyncio.create_task(self.upload_cloud())

    async def poll(self):
        log.info("tasks: starting poll ...")

        # if not os.path.exists(self.settings.processed):
        #     log.warning("create %s folder", self.settings.processed)
        #     os.makedirs(self.settings.processed)
        gps_started = False
        while True:
            if self.gps.time_valid and not gps_started:
                gps_started = True
                log.info("gps is valid, now notify systemd to start dcam with flag %s", self.settings.flag_gps_fix)
                with open(self.settings.flag_gps_fix, 'w') as flag: # watched by systemctl to start dcam.service
                    flag.write('fix')
            await asyncio.sleep(1)
#            self.publish("prova.sub", msg="mi leggi?")
            await self.localizePlates()
            await self.filterPlates(validate_delay=self.settings.validate_delay)

    async def upload_cloud(self, state_initial=2, state_final=3):
        """
        da stato 2 a stato 3 se spedite correttamente
        """

        # def sliceIterator(lst, sliceLen):
        #     for i in range(len(lst) - sliceLen + 1):
        #         yield lst[i:i + sliceLen]
        def sliceIterator(l, n):  # seconda versione
            for i in range(0, len(l), n):
                yield l[i : i + n]

        api = api_cloud.Api(config=self.settings.api, mev_unit=self.settings.mev_unit)
        if not api.enabled():
            log.warning("api cloud not enabled, skip running upload_cloud")
            return

        while True:
            try:
                if not api.login:
                    login = await api.extDeviceSignClient()
                    if not login:
                        log.warning("mev_cloud: cannot login")
                        await asyncio.sleep(10)
                        continue
                    self.api = api

                plates = self.db_plates.get(state=state_initial)
                if plates:
                    # mandiamo un massimo di 30 targhe
                    for sliceplates in sliceIterator(plates, min(len(plates), 30)):
                        uploaded = await api.recognizedPlatesClient(sliceplates)
                        # log.info("\033[32m" + f"uploaded {len(plates)} plates" + "\033[0m")
                        if uploaded:
                            log.info(
                                "mev_cloud uploaded %d plates: %s ...",
                                len(sliceplates),
                                [p["id"] for p in sliceplates[:3]],
                            )
                            self.publish(
                                "plate.uploaded", plates=[dict(p) for p in sliceplates]
                            )
                            for plate in sliceplates:
                                self.db_plates.updateState(plate, state_final)
                            self.db_plates.commit()
                else:
                    log.info("mev_cloud no plates to upload")
                await asyncio.sleep(10)
            except Exception as e:
                log.error("%s: retry in 10 seconds...", str(e))
                await asyncio.sleep(10)
                continue

    async def publishLocation(self, p2=None, **kwargs):
        log.info("publishLocation - gps status: ...")
        #TODO: aggiungere random speed o fixed speed (solo nel caso di debug)
        if self.gps.time_valid:
            self._counter_gps += 1
            location = {
                "latitude": p2.latitude,
                "longitude": p2.longitude,
                "speed": p2.speed,
                "satellites": p2.satellites,
                "hdop": p2.horizontal_dilution,
                "elevation": p2.elevation,
                "sensors": sensors.getStats(),
            }
            self.publish("gps.current", unit=self.settings.mev_unit, location=location)
            if self._counter_gps % 10 == 0: 
                # aggiorniamo la centrale solo una volta ogni 10 secondi poiché è remota
                if self.api:
                    self.publish("gps.unit", unit_id=self.api.unit_id, location=location)
                log.info("GPS: %s", location)
        else:
            self.publish("gps.status", message="not ready")

    def close(self):
        log.warning("closing database")
        self.db_plates.close()
        self.gps.close()

    async def localizePlates(self, state_initial=0, state_final=1):
        """
        setta latitudine e longitudine dei plates in stato 0
        """

        for plate in self.db_plates.get(state=state_initial):
            # last = self.db_plates.getLatest()
            # for i in range(0, 19):
            #     log.info("last plate " + str(last[i]))
            location = await self.gps.getLocationFromTime(
                timestamp=plate["timestamp"], quarter=plate["quarter"]
            )
            if location:
                self.publish(
                    "gps.marker",
                    location=dict(
                        latitude=location.latitude, longitude=location.longitude
                    ),
                )
                log.info(
                    "plate localized: ascii=%s quarter=%s location=%s",
                    plate["plate_ascii"],
                    plate["quarter"],
                    location,
                )
                self.db_plates.updateLocation(plate, location, state_final)
            else:
                log.warning(
                    "no location for plate ascii=%s timestamp=%f quarter=%f",
                    plate["plate_ascii"],
                    plate["timestamp"],
                    plate["quarter"],
                )

    def updateZone(self, plate):
        zones = self.kml.getZones(Location(plate["latitude"], plate["longitude"]))
        self.db_plates.updateZone(plate, zones)
        plate["zone"] = zones
        return zones

    def checkPlateInZone(self, plate, zones):
        """
        torna True se la targa è dentro una zona valida
        """

        if zones == 0:
            return False

        # escludiamo questi falsi positivi di macchine che stanno uscendo da un incrocio o un garage
        if plate["parking"] == ParkingAngle.vertical and not (
            zones & ZoneType.vertical_parking
        ):  # un plate verticale in una zona diversa da vertical_parking non è valido
            return False

        if zones & ZoneType.both:
            return True

        if (zones & ZoneType.right) and plate["quarter"] <= 1:
            return True

        return False

    async def filterPlates(
        self, max_distance=300, validate_delay=10, state_initial=1, state_final=2
    ):
        """
        - elimina le targhe doppie riconosciute da dcam0 e dcam1 tenendo le migliori.
         considera le targhe con un timestamp più vecchio di now di almeno validate_delay
        - localizza latitudine e longitudine nella zona
        """
        processed_plates = []
        for plate in self.db_plates.getOlder(
            state=state_initial, validate_delay=validate_delay
        ):
            if plate["key"] in processed_plates:
                continue

            similar_plates = self.db_plates.getSimilar(
                plate["plate_ascii"], max_distance=max_distance, state=state_initial
            )
            best = True
            for p in similar_plates:
                if best:
                    plate_best = dict(p)
                    state_for_best = state_final
                    zones = self.updateZone(plate_best)

                    # if '37006' in plate['plate_ascii']:
                    #     plate_best['parking'] = ParkingAngle.vertical

                    isValidZone = self.checkPlateInZone(plate_best, zones)
                    # if zone == ZoneType.off or \
                    #         (zone == ZoneType.right and plate_best['quarter'] > 1):
                    if not isValidZone:
                        state_for_best = -state_final - 1  # escludiamo dall'upload
                    plate_best["state"] = state_for_best
                    self.db_plates.updateState(plate_best, state_for_best)
                    self.publish("plate.valid", plate=plate_best)
                    log.info(
                        "plate detected: plate=%s state=%s zones=%s quarter=%s parking=%s",
                        plate_best["plate"],
                        plate_best["state"],
                        zones,
                        plate_best["quarter"],
                        plate_best["parking"],
                    )
                    best = False  # solo la prima ha lo score maggiore
                    if self.settings.plates_truth:
                        self.truth_plates.add(plate_best["plate_ascii"])
                else:
                    self.db_plates.updateState(
                        p, -state_final
                    )  # scartiamo questa targa

                processed_plates.append(p["key"])

    async def saveToggle(self, status):
        log.info("saveToggle(%s)", status)
        if status:
            utils.sendSignalToProcesses(
                "dcam_run", sig="SIGUSR1"
            )  # salviamo file video mandando un segnale SIGUSR1 ai processi dcam_run
            await self.gps.track()
        else:
            utils.sendSignalToProcesses(
                "dcam_run", sig="SIGUSR2"
            )  # interrompiamo la registrazione
            await self.gps.stop()
        self.save_status = status

    def getStatusSaving(self):
        return self.save_status

    @Authorize()
    async def getContextImage(self, request, audience="", payload={}, key=""):
        resp = await self.getPlateImage(
            key=key, size=None, for_http=False, add_crop=True
        )
        return response.json(resp)

    @Authorize()
    async def getImage(
        self, request, audience="", payload={}, key="", kind="context", encode="binary"
    ):
        """
        TODO encode==base64 non implementato
        """
        if key == "latest":
            plate = self.db_plates.getLatest()
            if plate:
                key = plate["key"]
            else:
                return response.text("")
        _bytes = await self.getPlateImage(
            key=key,
            size=None,
            binary=encode == "binary",
            for_http=False,
            add_crop=True,
            kind=kind,
        )
        return response.raw(_bytes, headers={"Content-type": "image/jpeg"})

    async def onDisconnect(self):
        log.warning("wamp transport lost, shutting down")
        self.close()
        asyncio.get_event_loop().stop()

    async def onPlate(self, plate=None):
        """
        se il GPS time non è corretto non inseriamo le targhe
        """
        log.info(f"sono in brain per il subscribe {plate}")
        if not self.gps.time_valid:
            log.warning("GPS time not ready, skip new plate detected")
            return
        # if not ("DK" in plate['plate_ascii']):
        #    return
        self.db_plates.add(**plate)

    async def onSaveToggle(self, status=False):
        await self.saveToggle(status)

    async def onGpsRemote(self, *args, **kwargs):
        log.info("got %s", kwargs)
        await self.gps.addPointFromRemote(kwargs["coords"], kwargs["timestamp"])

    async def getStatus(self, *args, **kwargs):
        return {"saving": self.getStatusSaving()}

    async def getPlateImage(
        self,
        key="",
        size=(100, 50),
        binary=False,
        for_http=True,
        add_crop=False,
        kind="context",
        size_overlay_perc=30,
    ):
        # plate = self.db_plates.getByKey(key)
        # if not plate:
        #    log.error(f'plate key={key} not present in db')
        #    return {'base64':''}

        filename_context = os.path.join(self.settings.plate_images, "%s.jpg" % key)
        filename_crop = os.path.join(self.settings.plate_images, "%s_plate.jpg" % key)
        log.debug("getPlateimage key=%s", key)
        if not os.path.exists(filename_context):
            log.warning("getPlateimage image %s not found", filename_context)
            data = ""
        else:
            img = cv.imread(filename_context if kind == "context" else filename_crop)
            if kind == "context" and add_crop and os.path.exists(filename_crop):
                label = os.path.basename(filename_context).split("_")[0]

                overlay = cv.imread(filename_crop)
                if size_overlay_perc > 0:
                    h,w = img.shape[:2]
                    w_overlay = int(w * size_overlay_perc / 100)
                    ratio = float(w_overlay)/overlay.shape[1]
                    h_overlay = int(overlay.shape[0]*ratio)
                    overlay = cv.resize(overlay, (w_overlay, h_overlay), interpolation = cv.INTER_AREA)
                utils.add_overlay(
                    img, overlay, label=label, fontSize=1.0
                )  # , label2=str(plate['angle']))

            if binary:
                return utils.toBinary(img)
            data = utils.getBase64(img, size, for_http=for_http)
        return {"base64": data}

    async def getZones(self):
        with open(self.settings.zones) as f:
            return {"kml": f.read()}


def main():
    config = initParser()
    if config.debug:
        import jurigged

        jurigged.watch(pattern="*mev_brain*.py")

    poller = PollerMev(config)
    try:
        asyncio.get_event_loop().run_forever()
    except KeyboardInterrupt:
        poller.close()
    finally:
        print("Bye")
